# Orca
#
# Copyright 2004-2008 Sun Microsystems Inc.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., Franklin Street, Fifth Floor,
# Boston MA  02110-1301 USA.

# pylint: disable=too-many-public-methods

"""Custom script for Thunderbird."""

from __future__ import annotations

__id__        = "$Id$"
__version__   = "$Revision$"
__date__      = "$Date$"
__copyright__ = "Copyright (c) 2004-2008 Sun Microsystems Inc."
__license__   = "LGPL"

from typing import TYPE_CHECKING

from orca import cmdnames
from orca import debug
from orca import focus_manager
from orca import input_event
from orca import settings_manager
from orca.scripts.toolkits import Gecko
from orca.ax_object import AXObject
from orca.ax_text import AXText
from orca.ax_utilities import AXUtilities

from .spellcheck import SpellCheck

if TYPE_CHECKING:
    import gi
    gi.require_version("Atspi", "2.0")
    gi.require_version("Gtk", "3.0")
    from gi.repository import Atspi, Gtk

    from orca.scripts.toolkits.Gecko.script_utilities import Utilities

class Script(Gecko.Script):
    """The script for Thunderbird."""

    # Override the base class type annotations
    spellcheck: SpellCheck
    utilities: "Utilities"

    def __init__(self, app: Atspi.Accessible) -> None:
        if settings_manager.get_manager().get_setting('sayAllOnLoad') is None:
            settings_manager.get_manager().set_setting('sayAllOnLoad', False)
        if settings_manager.get_manager().get_setting('pageSummaryOnLoad') is None:
            settings_manager.get_manager().set_setting('pageSummaryOnLoad', False)

        super().__init__(app)

    def setup_input_event_handlers(self) -> None:
        """Defines the input event handlers for this script."""

        super().setup_input_event_handlers()

        self.input_event_handlers["togglePresentationModeHandler"] = \
            input_event.InputEventHandler(
                Script.toggle_presentation_mode,
                cmdnames.TOGGLE_PRESENTATION_MODE)

        self.input_event_handlers["enableStickyFocusModeHandler"] = \
            input_event.InputEventHandler(
                Script.enable_sticky_focus_mode,
                cmdnames.SET_FOCUS_MODE_STICKY)

        self.input_event_handlers["enableStickyBrowseModeHandler"] = \
            input_event.InputEventHandler(
                Script.enable_sticky_browse_mode,
                cmdnames.SET_BROWSE_MODE_STICKY)

    def get_spellcheck(self) -> SpellCheck:
        """Returns the spellcheck support for this script."""

        return SpellCheck(self)

    def get_app_preferences_gui(self) -> Gtk.Grid:
        """Return a GtkGrid containing the application unique configuration
        GUI items for the current application."""

        grid = super().get_app_preferences_gui()

        assert self._say_all_on_load_check_button is not None
        assert self._page_summary_on_load_check_button is not None
        self._say_all_on_load_check_button.set_active(
            settings_manager.get_manager().get_setting("sayAllOnLoad"))
        self._page_summary_on_load_check_button.set_active(
            settings_manager.get_manager().get_setting("pageSummaryOnLoad"))

        spellcheck = self.spellcheck.get_app_preferences_gui()
        grid.attach(spellcheck, 0, len(grid.get_children()), 1, 1)
        grid.show_all()

        return grid

    def get_preferences_from_gui(self) -> dict[str, bool | int]:
        """Returns a dictionary with the app-specific preferences."""

        prefs = super().get_preferences_from_gui()
        assert self._say_all_on_load_check_button is not None
        assert self._page_summary_on_load_check_button is not None
        prefs["sayAllOnLoad"] = self._say_all_on_load_check_button.get_active()
        prefs["pageSummaryOnLoad"] = self._page_summary_on_load_check_button.get_active()
        prefs.update(self.spellcheck.get_preferences_from_gui())

        return prefs

    def locus_of_focus_changed(
        self,
        event: Atspi.Event | None,
        old_focus: Atspi.Accessible | None,
        new_focus: Atspi.Accessible | None
    ) -> bool:
        """Handles changes of focus of interest. Returns True if this script did all needed work."""

        if self.spellcheck.is_suggestions_item(new_focus):
            include_label = not self.spellcheck.is_suggestions_item(old_focus)
            self.update_braille(new_focus)
            self.spellcheck.present_suggestion_list_item(include_label=include_label)
            return True

        return super().locus_of_focus_changed(event, old_focus, new_focus)

    def use_focus_mode(
        self,
        obj: Atspi.Accessible,
        prev_obj: Atspi.Accessible | None = None
    ) -> bool:
        if self.utilities.is_editable_message(obj):
            tokens = ["THUNDERBIRD: Using focus mode for editable message", obj]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return True

        tokens = ["THUNDERBIRD:", obj, "is not an editable message."]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)
        return super().use_focus_mode(obj, prev_obj)

    def enable_sticky_browse_mode(
        self,
        event: input_event.InputEvent | None = None,
        force_message: bool = False
    ) -> bool:
        if self.utilities.is_editable_message(focus_manager.get_manager().get_locus_of_focus()):
            return True

        return super().enable_sticky_browse_mode(event, force_message)

    def enable_sticky_focus_mode(
        self,
        event: input_event.InputEvent | None = None,
        force_message: bool = False
    ) -> bool:
        if self.utilities.is_editable_message(focus_manager.get_manager().get_locus_of_focus()):
            return True

        return super().enable_sticky_focus_mode(event, force_message)

    def toggle_presentation_mode(
        self,
        event: input_event.InputEvent | None = None,
        document: Atspi.Accessible | None = None,
        notify_user: bool = True
    ) -> bool:
        if self._in_focus_mode \
           and self.utilities.is_editable_message(focus_manager.get_manager().get_locus_of_focus()):
            return True

        return super().toggle_presentation_mode(event, document)

    def on_busy_changed(self, event: Atspi.Event) -> bool:
        """Callback for object:state-changed:busy accessibility events."""

        # TODO - JD: Can this logic be moved to the web script?
        if self.utilities.is_editable_message(event.source):
            return True

        if self.in_focus_mode():
            return True

        return super().on_busy_changed(event)

    def on_caret_moved(self, event: Atspi.Event) -> bool:
        """Callback for object:text-caret-moved accessibility events."""

        if self.utilities.is_editable_message(event.source):
            if event.detail1 == -1:
                return True
            self.spellcheck.set_document_position(event.source, event.detail1)
            if self.spellcheck.is_active():
                return True

        return super().on_caret_moved(event)

    def on_name_changed(self, event: Atspi.Event) -> bool:
        """Callback for object:property-change:accessible-name events."""

        if AXObject.get_name(event.source) == self.spellcheck.get_misspelled_word():
            self.spellcheck.present_error_details()
            return True

        return super().on_name_changed(event)

    def on_selection_changed(self, event: Atspi.Event) -> bool:
        """Callback for object:state-changed:showing accessibility events."""

        # We present changes when the list has focus via focus-changed events.
        if event.source == self.spellcheck.get_suggestions_list():
            return True

        parent = AXObject.get_parent(event.source)
        if AXUtilities.is_combo_box(parent) and not AXUtilities.is_focused(parent):
            return True

        return super().on_selection_changed(event)

    def on_sensitive_changed(self, event: Atspi.Event) -> bool:
        """Callback for object:state-changed:sensitive accessibility events."""

        if event.source == self.spellcheck.get_change_to_entry() \
           and self.spellcheck.present_completion_message():
            return True

        return super().on_sensitive_changed(event)

    def on_text_deleted(self, event: Atspi.Event) -> bool:
        """Callback for object:text-changed:delete accessibility events."""

        if AXUtilities.is_label(event.source) \
           and AXUtilities.is_status_bar(AXObject.get_parent(event.source)):
            return True

        return super().on_text_deleted(event)

    def on_text_inserted(self, event: Atspi.Event) -> bool:
        """Callback for object:text-changed:insert accessibility events."""

        parent = AXObject.get_parent(event.source)
        if AXUtilities.is_label(event.source) and AXUtilities.is_status_bar(parent):
            return True

        if len(event.any_data) > 1 and event.source == self.spellcheck.get_change_to_entry():
            return True

        # Try to stop unwanted chatter when a message is being replied to.
        # See bgo#618484.
        if event.type.endswith("system") and self.utilities.is_editable_message(event.source):
            return True

        return super().on_text_inserted(event)

    def on_text_selection_changed(self, event: Atspi.Event) -> bool:
        """Callback for object:text-selection-changed accessibility events."""

        if event.source == self.spellcheck.get_change_to_entry():
            return True

        _reason = AXUtilities.get_text_event_reason(event)
        if self.utilities.is_editable_message(event.source) and self.spellcheck.is_active():
            selection_start = AXText.get_selection_start_offset(event.source)
            if selection_start >= 0:
                self.spellcheck.set_document_position(event.source, selection_start)
            return True

        return super().on_text_selection_changed(event)

    def on_window_activated(self, event: Atspi.Event) -> bool:
        """Callback for window:activate accessibility events."""

        super().on_window_activated(event)
        if not self.spellcheck.is_spell_check_window(event.source):
            self.spellcheck.deactivate()
            return True

        self.spellcheck.present_error_details()
        entry = self.spellcheck.get_change_to_entry()
        focus_manager.get_manager().set_locus_of_focus(None, entry, False)
        self.update_braille(entry)
        return True

    def on_window_deactivated(self, event: Atspi.Event) -> bool:
        """Callback for window:deactivate accessibility events."""

        self.spellcheck.deactivate()
        self.utilities.clear_content_cache()
        return super().on_window_deactivated(event)



# Orca
#
# Copyright 2025 Valve Corporation
# Author: Joanmarie Diggs <jdiggs@igalia.com>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., Franklin Street, Fifth Floor,
# Boston MA  02110-1301 USA.

# pylint: disable=too-few-public-methods
# pylint: disable=too-many-return-statements
# pylint: disable=too-many-instance-attributes
# pylint: disable=too-many-branches
# pylint: disable=too-many-locals
# pylint: disable=too-many-nested-blocks
# pylint: disable=too-many-arguments
# pylint: disable=too-many-positional-arguments
# pylint: disable=too-many-statements

"""Provides a D-Bus interface for remotely controlling Orca."""

__id__        = "$Id$"
__version__   = "$Revision$"
__date__      = "$Date$"
__copyright__ = "Copyright (c) 2025 Valve Corporation."
__license__   = "LGPL"

import enum
import inspect
from typing import Callable

from dasbus.connection import SessionMessageBus
from dasbus.error import DBusError
from dasbus.loop import EventLoop
from dasbus.server.interface import dbus_interface
from dasbus.server.publishable import Publishable
from gi.repository import GLib

from . import debug
from . import input_event
from . import input_event_manager
from . import orca_platform # pylint: disable=no-name-in-module
from . import script_manager

class HandlerType(enum.Enum):
    """Enumeration of handler types for D-Bus methods."""

    COMMAND = enum.auto()
    PARAMETERIZED_COMMAND = enum.auto()
    GETTER = enum.auto()
    SETTER = enum.auto()

def command(func):
    """Decorator to mark a method as a D-Bus command using its docstring.

    Usage:
        @command
        def toggle_speech(self, script=None, event=None):
            '''Toggles speech on and off.'''
            # method implementation
    """
    description = func.__doc__ or f"D-Bus command: {func.__name__}"
    func.dbus_command_description = description
    return func

def parameterized_command(func):
    """Decorator to mark a method as a D-Bus parameterized command using its docstring.

    Usage:
        @parameterized_command
        def get_voices_for_language(
            self,
            language,
            variant='',
            script=None,
            event=None,
            notify_user=False
        ):
            '''Returns a list of available voices for the specified language.'''
            # method implementation
    """
    description = func.__doc__ or f"D-Bus parameterized command: {func.__name__}"
    func.dbus_parameterized_command_description = description
    return func

def getter(func):
    """Decorator to mark a method as a D-Bus getter using its docstring.

    Usage:
        @getter
        def get_rate(self):
            '''Returns the current speech rate.'''
            # method implementation
    """
    description = func.__doc__ or f"D-Bus getter: {func.__name__}"
    func.dbus_getter_description = description
    return func

def setter(func):
    """Decorator to mark a method as a D-Bus setter using its docstring.

    Usage:
        @setter
        def set_rate(self, value):
            '''Sets the current speech rate.'''
            # method implementation
    """
    description = func.__doc__ or f"D-Bus setter: {func.__name__}"
    func.dbus_setter_description = description
    return func


def _extract_function_parameters(func: Callable) -> list[tuple[str, str]]:
    """Extract parameter names and types from a function signature."""

    sig = inspect.signature(func)
    parameters = []

    skip_params = {"self", "script", "event"}
    for param_name, param in sig.parameters.items():
        if param_name in skip_params:
            continue

        if param.annotation != inspect.Parameter.empty:
            if hasattr(param.annotation, "__name__"):
                type_str = param.annotation.__name__
            else:
                type_str = str(param.annotation).replace("typing.", "")
        else:
            type_str = "Any"
        parameters.append((param_name, type_str))

    return parameters


class _HandlerInfo:
    """Stores processed information about a function exposed via D-Bus."""

    def __init__(
        self,
        python_function_name: str,
        description: str,
        action: Callable[..., bool],
        handler_type: HandlerType = HandlerType.COMMAND,
        parameters: list[tuple[str, str]] | None = None
    ):
        self.python_function_name: str = python_function_name
        self.description: str = description
        self.action: Callable[..., bool] = action
        self.handler_type: HandlerType = handler_type
        self.parameters: list[tuple[str, str]] = parameters or []


@dbus_interface("org.gnome.Orca.Module")
class OrcaModuleDBusInterface(Publishable):
    """A D-Bus interface representing a specific Orca module (e.g., a manager)."""

    def __init__(self,
                 module_name: str,
                 handlers_info: list[_HandlerInfo]):
        super().__init__()
        self._module_name = module_name
        self._commands: dict[str, _HandlerInfo] = {}
        self._parameterized_commands: dict[str, _HandlerInfo] = {}
        self._getters: dict[str, _HandlerInfo] = {}
        self._setters: dict[str, _HandlerInfo] = {}

        for info in handlers_info:
            handler_type = getattr(info, "handler_type", HandlerType.COMMAND)
            normalized_name = self._normalize_handler_name(info.python_function_name, handler_type)
            if handler_type == HandlerType.GETTER:
                self._getters[normalized_name] = info
            elif handler_type == HandlerType.SETTER:
                self._setters[normalized_name] = info
            elif handler_type == HandlerType.PARAMETERIZED_COMMAND:
                self._parameterized_commands[normalized_name] = info
            else:
                self._commands[normalized_name] = info

        msg = (
            f"DBUS SERVICE: OrcaModuleDBusInterface for {module_name} initialized "
            f"with {len(self._commands)} command(s), "
            f"{len(self._parameterized_commands)} parameterized command(s), "
            f"{len(self._getters)} getter(s), {len(self._setters)} setter(s)."
        )
        debug.print_message(debug.LEVEL_INFO, msg, True)

    def ExecuteRuntimeGetter(self, getter_name: str) -> GLib.Variant: # pylint: disable=invalid-name
        """Executes the named getter returning the value as a GLib.Variant for D-Bus marshalling."""

        handler_info = self._getters.get(getter_name)
        if not handler_info:
            msg = f"DBUS SERVICE: Unknown getter '{getter_name}' for '{self._module_name}'."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return GLib.Variant("v", GLib.Variant("s", ""))

        result = handler_info.action()
        msg = f"DBUS SERVICE: Getter '{getter_name}' returned: {result}"
        debug.print_message(debug.LEVEL_INFO, msg, True)
        return self._to_variant(result)

    def ExecuteRuntimeSetter(self, setter_name: str, value: GLib.Variant) -> bool: # pylint: disable=invalid-name
        """Executes the named setter, returning True if succeeded."""

        handler_info = self._setters.get(setter_name)
        if handler_info is None:
            msg = f"DBUS SERVICE: Unknown setter '{setter_name}' for '{self._module_name}'."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return False

        unpacked = value.unpack()
        result = handler_info.action(unpacked)
        msg = f"DBUS SERVICE: Setter '{setter_name}' with value '{unpacked}' returned: {result}"
        debug.print_message(debug.LEVEL_INFO, msg, True)
        return result

    def ListCommands(self) -> list[tuple[str, str]]: # pylint: disable=invalid-name
        """Returns a list of (command_name, description) for this module (commands only)."""

        command_list = []
        for camel_case_name, info in self._commands.items():
            command_list.append((camel_case_name, info.description))
        return command_list

    def ListParameterizedCommands(  # pylint: disable=invalid-name
        self,
    ) -> list[tuple[str, str, list[tuple[str, str]]]]:
        """Returns a list of (command_name, description, parameters) for this module."""

        command_list = []
        for camel_case_name, info in self._parameterized_commands.items():
            command_list.append((camel_case_name, info.description, info.parameters))
        return command_list

    def ListRuntimeGetters(self) -> list[tuple[str, str]]: # pylint: disable=invalid-name
        """Returns a list of (getter_name, description) for this module."""

        getter_list = []
        for camel_case_name, info in self._getters.items():
            getter_list.append((camel_case_name, info.description))
        return getter_list

    def ListRuntimeSetters(self) -> list[tuple[str, str]]: # pylint: disable=invalid-name
        """Returns a list of (setter_name, description) for this module."""

        setter_list = []
        for camel_case_name, info in self._setters.items():
            setter_list.append((camel_case_name, info.description))
        return setter_list

    def ExecuteCommand(self, command_name: str, notify_user: bool) -> bool: # pylint: disable=invalid-name
        """Executes the named command and returns True if the command succeeded."""

        if command_name not in self._commands:
            msg = f"DBUS SERVICE: Unknown command '{command_name}' for '{self._module_name}'."
            debug.print_message(debug.LEVEL_WARNING, msg, True)
            return False

        handler_info = self._commands[command_name]
        msg = f"DBUS SERVICE: About to execute '{command_name}' in '{self._module_name}'."
        debug.print_message(debug.LEVEL_INFO, msg, True)
        result = handler_info.action(notify_user)
        msg = (
            f"DBUS SERVICE: '{command_name}' in '{self._module_name}' executed. "
            f"Result: {result}, notify_user: {notify_user}"
        )
        debug.print_message(debug.LEVEL_INFO, msg, True)
        return result

    def ExecuteParameterizedCommand( # pylint: disable=invalid-name
        self,
        command_name: str,
        parameters: dict[str, GLib.Variant],
        notify_user: bool
    ) -> GLib.Variant:
        """Executes the named command with parameters and returns the result."""

        handler_info = self._parameterized_commands.get(command_name)
        if not handler_info:
            msg = (
                f"DBUS SERVICE: Unknown parameterized command '{command_name}' for "
                f"'{self._module_name}'."
            )
            debug.print_message(debug.LEVEL_WARNING, msg, True)
            return GLib.Variant("b", False)

        kwargs = {name: variant.unpack() for name, variant in parameters.items()}
        kwargs["notify_user"] = notify_user
        msg = f"DBUS SERVICE: About to execute '{command_name}' in '{self._module_name}'."
        debug.print_message(debug.LEVEL_INFO, msg, True)
        result = handler_info.action(**kwargs)
        msg = f"DBUS SERVICE: '{command_name}' in '{self._module_name}' executed."
        debug.print_message(debug.LEVEL_INFO, msg, True)
        return self._to_variant(result)

    def for_publication(self):
        """Returns the D-Bus interface XML for publication."""

        return self.__dbus_xml__ # pylint: disable=no-member

    @staticmethod
    def _normalize_handler_name(
        function_name: str,
        handler_type: HandlerType = HandlerType.COMMAND
    ) -> str:
        """Normalizes a Python function name for D-Bus exposure (getter/setter/command)."""

        # Only strip prefixes for getters and setters, not for commands
        if handler_type in (HandlerType.GETTER, HandlerType.SETTER):
            if function_name.startswith("get_") or function_name.startswith("set_"):
                function_name = function_name[4:]
        return "".join(word.capitalize() for word in function_name.split("_"))

    @staticmethod
    def _to_variant(result):
        """Converts a Python value to a correctly-typed GLib.Variant for D-Bus marshalling."""

        if isinstance(result, bool):
            return GLib.Variant("b", result)
        if isinstance(result, int):
            return GLib.Variant("i", result)
        if isinstance(result, float):
            return GLib.Variant("d", result)
        if isinstance(result, str):
            return GLib.Variant("s", result)
        if isinstance(result, dict):
            return GLib.Variant(
                "a{sv}", {str(k): GLib.Variant("v", v) for k, v in result.items()})
        if isinstance(result, (list, tuple)):
            if all(isinstance(x, str) for x in result):
                return GLib.Variant("as", list(result))
            if all(isinstance(x, bool) for x in result):
                return GLib.Variant("ab", list(result))
            if all(isinstance(x, int) for x in result):
                return GLib.Variant("ax", list(result))
            if all(isinstance(x, (list, tuple)) for x in result):
                if not result:
                    return GLib.Variant("av", [])
                first_len = len(result[0])
                converted = [tuple(str(item or "") for item in x) for x in result]
                signature = "(" + "s" * first_len + ")"
                return GLib.Variant(f"a{signature}", converted)
            return GLib.Variant("av", [GLib.Variant("v", x) for x in result])
        if result is None:
            return GLib.Variant("v", GLib.Variant("s", ""))
        return GLib.Variant("s", str(result))


@dbus_interface("org.gnome.Orca.Service")
class OrcaDBusServiceInterface(Publishable):
    """Internal D-Bus service object that handles D-Bus specifics."""

    def __init__(self) -> None:
        super().__init__()
        self._registered_modules: set[str] = set()

    def for_publication(self):
        """Returns the D-Bus interface XML for publication."""

        return self.__dbus_xml__  # pylint: disable=no-member

    def add_module_interface(
        self,
        module_name: str,
        handlers_info: list[_HandlerInfo],
        bus: SessionMessageBus,
        object_path_base: str
    ) -> None:
        """Creates and prepares a D-Bus interface for an Orca module."""

        object_path = f"{object_path_base}/{module_name}"
        if module_name in self._registered_modules:
            msg = f"DBUS SERVICE: Interface {module_name} already registered. Replacing."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            try:
                bus.unpublish_object(object_path)
            except DBusError as e:
                msg = f"DBUS SERVICE: Error unpublishing old interface for {module_name}: {e}"
                debug.print_message(debug.LEVEL_INFO, msg, True)
            self._registered_modules.discard(module_name)
        try:
            module_iface = OrcaModuleDBusInterface(module_name, handlers_info)
            bus.publish_object(object_path, module_iface)
            self._registered_modules.add(module_name)
            msg = f"DBUS SERVICE: Successfully published {module_name} at {object_path}."
            debug.print_message(debug.LEVEL_INFO, msg, True)
        except DBusError as e:
            msg = (
                f"DBUS SERVICE: Failed to create or publish D-Bus interface for "
                f"module {module_name} at {object_path}: {e}"
            )
            debug.print_message(debug.LEVEL_SEVERE, msg, True)

    def remove_module_interface(
        self,
        module_name: str,
        bus: SessionMessageBus,
        object_path_base: str
    ) -> bool:
        """Removes and unpublishes a D-Bus interface for an Orca module."""

        if module_name not in self._registered_modules:
            msg = f"DBUS SERVICE: Module {module_name} is not registered."
            debug.print_message(debug.LEVEL_WARNING, msg, True)
            return False

        object_path = f"{object_path_base}/{module_name}"
        try:
            bus.unpublish_object(object_path)
            self._registered_modules.discard(module_name)
            msg = f"DBUS SERVICE: Successfully removed {module_name} from {object_path}."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True
        except DBusError as e:
            msg = f"DBUS SERVICE: Error removing interface for {module_name}: {e}"
            debug.print_message(debug.LEVEL_WARNING, msg, True)
            return False

    def ListModules(self) -> list[str]: # pylint: disable=invalid-name
        """Returns a list of registered module names."""

        return list(self._registered_modules)

    def ListCommands(self) -> list[tuple[str, str]]: # pylint: disable=invalid-name
        """Returns available commands on the main service interface."""

        commands = []
        for attr_name in dir(self):
            if not attr_name.startswith("_") and attr_name[0].isupper():
                attr = getattr(self, attr_name)
                if callable(attr) and hasattr(attr, "__doc__"):
                    description = (attr.__doc__.strip() if attr.__doc__
                                 else f"Service command: {attr_name}")
                    commands.append((attr_name, description))

        return sorted(commands)

    def ShowPreferences(self) -> bool: # pylint: disable=invalid-name
        """Shows Orca's preferences GUI."""

        msg = "DBUS SERVICE: ShowPreferences called."
        debug.print_message(debug.LEVEL_INFO, msg, True)

        manager = script_manager.get_manager()
        script = manager.get_active_script() or manager.get_default_script()
        if script is None:
            msg = "DBUS SERVICE: No script available"
            debug.print_message(debug.LEVEL_WARNING, msg, True)
            return False

        script.show_preferences_gui()
        return True

    def PresentMessage(self, message: str) -> bool: # pylint: disable=invalid-name
        """Presents message to the user."""

        msg = f"DBUS SERVICE: PresentMessage called with: '{message}'"
        debug.print_message(debug.LEVEL_INFO, msg, True)

        manager = script_manager.get_manager()
        script = manager.get_active_script() or manager.get_default_script()
        if script is None:
            msg = "DBUS SERVICE: No script available"
            debug.print_message(debug.LEVEL_WARNING, msg, True)
            return False

        script.present_message(message)
        return True

    def GetVersion(self) -> str: # pylint: disable=invalid-name
        """Returns Orca's version and revision if available."""

        result = orca_platform.version
        if orca_platform.revision:
            result += f" (rev {orca_platform.revision})"

        msg = f"DBUS SERVICE: GetVersion called, returning: {result}"
        debug.print_message(debug.LEVEL_INFO, msg, True)
        return result

    def Quit(self) -> bool: # pylint: disable=invalid-name
        """Quits Orca. Returns True if the quit request was accepted."""

        msg = "DBUS SERVICE: Quit called."
        debug.print_message(debug.LEVEL_INFO, msg, True)

        from . import orca  # pylint: disable=import-outside-toplevel

        # orca.shutdown() shuts down the dbus service, so send the response immediately and then
        # do the actual shutdown after a brief delay.
        def _delayed_shutdown():
            orca.shutdown()
            return False

        GLib.timeout_add(100, _delayed_shutdown)
        return True

    def shutdown_service(self, bus: SessionMessageBus, object_path_base: str) -> None:
        """Releases D-Bus resources held by this service and its modules."""

        msg = "DBUS SERVICE: Releasing D-Bus resources for service."
        debug.print_message(debug.LEVEL_INFO, msg, True)

        for module_name in list(self._registered_modules):
            module_object_path = f"{object_path_base}/{module_name}"
            msg = (
                f"DBUS SERVICE: Shutting down and unpublishing module {module_name} "
                f"from main service."
            )
            debug.print_message(debug.LEVEL_INFO, msg, True)
            try:
                bus.unpublish_object(module_object_path)
            except DBusError as e:
                msg = f"DBUS SERVICE: Error unpublishing interface for {module_name}: {e}"
                debug.print_message(debug.LEVEL_INFO, msg, True)
        self._registered_modules.clear()

class OrcaRemoteController:
    """Manages Orca's D-Bus service for remote control."""

    SERVICE_NAME = "org.gnome.Orca.Service"
    OBJECT_PATH = "/org/gnome/Orca/Service"

    def __init__(self) -> None:
        self._dbus_service_interface: OrcaDBusServiceInterface | None = None
        self._is_running: bool = False
        self._bus: SessionMessageBus | None = None
        self._event_loop: EventLoop | None = None
        self._pending_registrations: dict[str, object] = {}
        self._total_commands: int = 0
        self._total_getters: int = 0
        self._total_setters: int = 0
        self._total_modules: int = 0

    def start(self) -> bool:
        """Starts the D-Bus service."""

        if self._is_running:
            msg = "REMOTE CONTROLLER: Start called but service is already running."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        msg = "REMOTE CONTROLLER: Attempting to start D-Bus service."
        debug.print_message(debug.LEVEL_INFO, msg, True)

        try:
            self._bus = SessionMessageBus()
            msg = (
                f"REMOTE CONTROLLER: SessionMessageBus acquired: "
                f"{self._bus.connection.get_unique_name()}"
            )
            debug.print_message(debug.LEVEL_INFO, msg, True)
        except DBusError as e:
            self._bus = None
            msg = f"REMOTE CONTROLLER: Failed to acquire D-Bus session bus: {e}"
            debug.print_message(debug.LEVEL_SEVERE, msg, True)
            return False

        self._dbus_service_interface = OrcaDBusServiceInterface()
        try:
            self._bus.publish_object(self.OBJECT_PATH, self._dbus_service_interface)
            self._bus.register_service(self.SERVICE_NAME)
        except DBusError as e:
            msg = f"REMOTE CONTROLLER: Failed to publish service or request name: {e}"
            debug.print_message(debug.LEVEL_SEVERE, msg, True)
            if self._dbus_service_interface and self._bus:
                try:
                    self._bus.unpublish_object(self.OBJECT_PATH)
                except DBusError:
                    pass
            self._dbus_service_interface = None
            self._bus = None
            return False

        self._is_running = True
        msg = (
            f"REMOTE CONTROLLER: Service started name={self.SERVICE_NAME} "
            f"path={self.OBJECT_PATH}."
        )
        debug.print_message(debug.LEVEL_INFO, msg, True)
        self._process_pending_registrations()
        self._print_registration_summary()
        return True

    def _process_pending_registrations(self) -> None:
        """Processes any module registrations that were queued before the service was ready."""

        if not self._pending_registrations:
            return

        msg = (
            f"REMOTE CONTROLLER: Processing {len(self._pending_registrations)} "
            f"pending module registrations."
        )
        debug.print_message(debug.LEVEL_INFO, msg, True)
        for module_name, module_instance in self._pending_registrations.items():
            msg = f"REMOTE CONTROLLER: Processing pending registration for {module_name}."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            self._register_decorated_commands_internal(module_name, module_instance)

        self._pending_registrations.clear()

    def register_decorated_module(self, module_name: str, module_instance) -> None:
        """Registers a module's decorated D-Bus commands."""

        if not self._is_running or not self._dbus_service_interface or not self._bus:
            msg = (
                f"REMOTE CONTROLLER: Service not ready; queuing decorated registration "
                f"for {module_name}."
            )
            debug.print_message(debug.LEVEL_INFO, msg, True)
            self._pending_registrations[module_name] = module_instance
            return

        self._register_decorated_commands_internal(module_name, module_instance)

    def _register_decorated_commands_internal(self, module_name: str, module_instance) -> None:
        """Internal method that registers decorated commands from a module instance."""

        if not self._is_running or not self._dbus_service_interface or not self._bus:
            msg = (
                f"REMOTE CONTROLLER: Internal error - _register_decorated_commands_internal "
                f"called for {module_name} but service is not ready."
            )
            debug.print_message(debug.LEVEL_SEVERE, msg, True)
            return

        handlers_info = []
        commands_count = 0
        getters_count = 0
        setters_count = 0

        for attr_name in dir(module_instance):
            attr = getattr(module_instance, attr_name)
            # Command
            if callable(attr) and hasattr(attr, "dbus_command_description"):
                description = attr.dbus_command_description
                def _create_wrapper(method=attr):
                    def _wrapper(notify_user):
                        event = input_event.RemoteControllerEvent()
                        script = script_manager.get_manager().get_active_script()
                        if script is None:
                            script = script_manager.get_manager().get_default_script()
                        rv = method(script=script, event=event, notify_user=notify_user)
                        # TODO - JD: It probably makes sense to fully process these input
                        # events just like any others, rather than have the caller here.
                        input_event_manager.get_manager().process_remote_controller_event(event)
                        return rv
                    return _wrapper
                handler_info = _HandlerInfo(
                    python_function_name=attr_name,
                    description=description,
                    action=_create_wrapper(),
                    handler_type=HandlerType.COMMAND
                )
                handlers_info.append(handler_info)
                commands_count += 1
            # Parameterized Command
            elif callable(attr) and hasattr(attr, "dbus_parameterized_command_description"):
                description = attr.dbus_parameterized_command_description
                def _create_parameterized_wrapper(method=attr):
                    def _wrapper(**kwargs):
                        event = input_event.RemoteControllerEvent()
                        script = script_manager.get_manager().get_active_script()
                        if script is None:
                            script = script_manager.get_manager().get_default_script()
                        rv = method(script=script, event=event, **kwargs)
                        # TODO - JD: It probably makes sense to fully process these input
                        # events just like any others, rather than have the caller here.
                        input_event_manager.get_manager().process_remote_controller_event(event)
                        return rv
                    return _wrapper
                handler_info = _HandlerInfo(
                    python_function_name=attr_name,
                    description=description,
                    action=_create_parameterized_wrapper(),
                    handler_type=HandlerType.PARAMETERIZED_COMMAND,
                    parameters=_extract_function_parameters(attr)
                )
                handlers_info.append(handler_info)
                commands_count += 1
            # Getter
            elif callable(attr) and hasattr(attr, "dbus_getter_description"):
                description = attr.dbus_getter_description
                def _create_getter_wrapper(method=attr):
                    def _wrapper(_notify_user=None):
                        return method()
                    return _wrapper
                handler_info = _HandlerInfo(
                    python_function_name=attr_name,
                    description=description,
                    action=_create_getter_wrapper(),
                    handler_type=HandlerType.GETTER
                )
                handlers_info.append(handler_info)
                getters_count += 1
            # Setter
            elif callable(attr) and hasattr(attr, "dbus_setter_description"):
                description = attr.dbus_setter_description
                def _create_setter_wrapper(method=attr):
                    def _wrapper(value):
                        return method(value)
                    return _wrapper
                handler_info = _HandlerInfo(
                    python_function_name=attr_name,
                    description=description,
                    action=_create_setter_wrapper(),
                    handler_type=HandlerType.SETTER
                )
                handlers_info.append(handler_info)
                setters_count += 1

        if not handlers_info:
            return

        self._total_commands += commands_count
        self._total_getters += getters_count
        self._total_setters += setters_count
        self._total_modules += 1

        self._dbus_service_interface.add_module_interface(
            module_name, handlers_info, self._bus, self.OBJECT_PATH)
        msg = (
            f"REMOTE CONTROLLER: Successfully registered {len(handlers_info)} "
            f"commands/getters/setters for module {module_name}."
        )
        debug.print_message(debug.LEVEL_INFO, msg, True)

    def deregister_module_commands(self, module_name: str) -> bool:
        """Deregisters D-Bus commands for an Orca module."""

        if module_name in self._pending_registrations:
            msg = f"REMOTE CONTROLLER: Removing pending registration for {module_name}."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            del self._pending_registrations[module_name]
            return True

        if not self._is_running or not self._dbus_service_interface or not self._bus:
            msg = (
                f"REMOTE CONTROLLER: Cannot deregister commands for {module_name}; "
                "service not running or bus not available."
            )
            debug.print_message(debug.LEVEL_WARNING, msg, True)
            return False

        return self._dbus_service_interface.remove_module_interface(
            module_name, self._bus, self.OBJECT_PATH)

    def shutdown(self) -> None:
        """Shuts down the D-Bus service."""

        if not self._is_running:
            msg = "REMOTE CONTROLLER: Shutdown called but service is not running."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return

        msg = "REMOTE CONTROLLER: Attempting to shut down D-Bus service."
        debug.print_message(debug.LEVEL_INFO, msg, True)
        if self._dbus_service_interface and self._bus:
            self._dbus_service_interface.shutdown_service(self._bus, self.OBJECT_PATH)
            try:
                self._bus.unpublish_object(self.OBJECT_PATH)
            except DBusError as e:
                msg = f"REMOTE CONTROLLER: Error unpublishing main service object: {e}"
                debug.print_message(debug.LEVEL_INFO, msg, True)
            self._dbus_service_interface = None

        if self._bus:
            try:
                self._bus.unregister_service(self.SERVICE_NAME)
            except DBusError as e:
                msg = f"REMOTE CONTROLLER: Error releasing bus name: {e}"
                debug.print_message(debug.LEVEL_INFO, msg, True)
            self._bus.disconnect()
            self._bus = None

        self._is_running = False
        msg = "REMOTE CONTROLLER: D-Bus service shut down."
        debug.print_message(debug.LEVEL_INFO, msg, True)
        self._pending_registrations.clear()
        self._total_commands = 0
        self._total_getters = 0
        self._total_setters = 0
        self._total_modules = 0

    def is_running(self) -> bool:
        """Checks if the D-Bus service is currently running."""

        return self._is_running

    def _count_system_commands(self) -> int:
        """Counts the system-wide D-Bus commands available on the main service interface."""

        if not self._dbus_service_interface:
            return 0

        system_commands = 0
        for attr_name in dir(self._dbus_service_interface):
            if not attr_name.startswith("_") and attr_name[0].isupper():
                attr = getattr(self._dbus_service_interface, attr_name)
                if callable(attr) and hasattr(attr, "__doc__"):
                    system_commands += 1
        return system_commands

    def _print_registration_summary(self) -> None:
        """Prints a summary of all registered D-Bus handlers."""

        system_commands_count = self._count_system_commands()
        total_handlers = self._total_commands + self._total_getters + self._total_setters
        msg = (
            f"REMOTE CONTROLLER: Registration complete. Summary: "
            f"{self._total_modules} modules, "
            f"{self._total_commands} module commands, "
            f"{self._total_getters} module getters, "
            f"{self._total_setters} module setters, "
            f"{system_commands_count} system commands. "
            f"Total handlers: {total_handlers + system_commands_count}."
        )
        debug.print_message(debug.LEVEL_INFO, msg, True)

_remote_controller: OrcaRemoteController = OrcaRemoteController()

def get_remote_controller() -> OrcaRemoteController:
    """Returns the OrcaRemoteController singleton."""

    return _remote_controller



# Orca
#
# Copyright 2023 The Orca Team
# Author: Rynhardt Kruger <rynkruger@gmail.com>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., Franklin Street, Fifth Floor,
# Boston MA  02110-1301 USA.

# pylint: disable=wrong-import-position

"""Provides ability to navigate objects hierarchically."""

# This has to be the first non-docstring line in the module to make linters happy.
from __future__ import annotations

__id__        = "$Id$"
__version__   = "$Revision$"
__date__      = "$Date$"
__copyright__ = "Copyright (c) 2023 The Orca Team"
__license__   = "LGPL"

from typing import TYPE_CHECKING

import gi
gi.require_version("Atspi", "2.0")
from gi.repository import Atspi

from . import cmdnames
from . import dbus_service
from . import debug
from . import focus_manager
from . import input_event
from . import keybindings
from . import messages
from .ax_event_synthesizer import AXEventSynthesizer
from .ax_object import AXObject
from .ax_utilities import AXUtilities

if TYPE_CHECKING:
    from .scripts import default

class ObjectNavigator:
    """Provides ability to navigate objects hierarchically."""

    def __init__(self) -> None:
        self._navigator_focus: Atspi.Accessible | None = None
        self._last_navigator_focus: Atspi.Accessible | None = None
        self._last_locus_of_focus: Atspi.Accessible | None = None
        self._simplify: bool = True
        self._handlers: dict = self.get_handlers(True)
        self._bindings: keybindings.KeyBindings = keybindings.KeyBindings()

        msg = "OBJECT NAVIGATOR: Registering D-Bus commands."
        debug.print_message(debug.LEVEL_INFO, msg, True)
        controller = dbus_service.get_remote_controller()
        controller.register_decorated_module("ObjectNavigator", self)

    def get_bindings(
        self, refresh: bool = False, is_desktop: bool = True
    ) -> keybindings.KeyBindings:
        """Returns the object-navigator keybindings."""

        if refresh:
            msg = f"OBJECT NAVIGATOR: Refreshing bindings. Is desktop: {is_desktop}"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            self._bindings.remove_key_grabs("OBJECT NAVIGATOR: Refreshing bindings.")
            self._setup_bindings()
        elif self._bindings.is_empty():
            self._setup_bindings()

        return self._bindings

    def get_handlers(self, refresh: bool = False) -> dict[str, input_event.InputEventHandler]:
        """Returns the object-navigator handlers."""

        if refresh:
            msg = "OBJECT NAVIGATOR: Refreshing handlers."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            self._setup_handlers()

        return self._handlers

    def _setup_bindings(self) -> None:
        """Sets up the object-navigator key bindings."""

        self._bindings = keybindings.KeyBindings()

        self._bindings.add(
            keybindings.KeyBinding(
                "Up",
                keybindings.DEFAULT_MODIFIER_MASK,
                keybindings.ORCA_CTRL_MODIFIER_MASK,
                self._handlers["object_navigator_up"]))

        self._bindings.add(
            keybindings.KeyBinding(
                "Down",
                keybindings.DEFAULT_MODIFIER_MASK,
                keybindings.ORCA_CTRL_MODIFIER_MASK,
                self._handlers["object_navigator_down"]))

        self._bindings.add(
            keybindings.KeyBinding(
                "Right",
                keybindings.DEFAULT_MODIFIER_MASK,
                keybindings.ORCA_CTRL_MODIFIER_MASK,
                self._handlers["object_navigator_next"]))

        self._bindings.add(
            keybindings.KeyBinding(
                "Left",
                keybindings.DEFAULT_MODIFIER_MASK,
                keybindings.ORCA_CTRL_MODIFIER_MASK,
                self._handlers["object_navigator_previous"]))

        self._bindings.add(
            keybindings.KeyBinding(
                "Return",
                keybindings.DEFAULT_MODIFIER_MASK,
                keybindings.ORCA_CTRL_MODIFIER_MASK,
                self._handlers["object_navigator_perform_action"]))

        self._bindings.add(
            keybindings.KeyBinding(
                "s",
                keybindings.DEFAULT_MODIFIER_MASK,
                keybindings.ORCA_CTRL_MODIFIER_MASK,
                self._handlers["object_navigator_toggle_simplify"]))

        msg = "OBJECT NAVIGATOR: Bindings set up."
        debug.print_message(debug.LEVEL_INFO, msg, True)

    def _setup_handlers(self) -> None:
        """Sets up the object-navigator input event handlers."""

        self._handlers = {}

        self._handlers["object_navigator_up"] = \
            input_event.InputEventHandler(
                self.move_to_parent,
                cmdnames.NAVIGATOR_UP)

        self._handlers["object_navigator_down"] = \
            input_event.InputEventHandler(
                self.move_to_first_child,
                cmdnames.NAVIGATOR_DOWN)

        self._handlers["object_navigator_next"] = \
            input_event.InputEventHandler(
                self.move_to_next_sibling,
                cmdnames.NAVIGATOR_NEXT)

        self._handlers["object_navigator_previous"] = \
            input_event.InputEventHandler(
                self.move_to_previous_sibling,
                cmdnames.NAVIGATOR_PREVIOUS)

        self._handlers["object_navigator_perform_action"] = \
            input_event.InputEventHandler(
                self.perform_action,
                cmdnames.NAVIGATOR_PERFORM_ACTION)

        self._handlers["object_navigator_toggle_simplify"] = \
            input_event.InputEventHandler(
                self.toggle_simplify,
                cmdnames.NAVIGATOR_TOGGLE_SIMPLIFIED)

        msg = "OBJECT NAVIGATOR: Handlers set up."
        debug.print_message(debug.LEVEL_INFO, msg, True)

    def _include_in_simple_navigation(self, obj: Atspi.Accessible) -> bool:
        """Returns True if obj should be included in simple navigation."""

        return AXUtilities.is_paragraph(obj)

    def _exclude_from_simple_navigation(
        self, _script: default.Script, obj: Atspi.Accessible
    ) -> bool:
        """Returns True if obj should be excluded from simple navigation."""

        if self._include_in_simple_navigation(obj):
            tokens = ["OBJECT NAVIGATOR: Not excluding", obj, ": explicit inclusion"]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return False

        # is_layout_only should catch things that really should be skipped.
        #
        # You do not want to exclude all sections because they may be focusable, e.g.
        # <div tabindex=0>foo</div> should not be excluded, despite the poor authoring.
        #
        # You do not want to exclude table cells and headers because it will make the
        # selectable items in tables non-navigable (e.g. the mail folders in Evolution)
        if AXUtilities.is_layout_only(obj):
            tokens = ["OBJECT NAVIGATOR: Excluding", obj, ": is layout only"]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return True

        tokens = ["OBJECT NAVIGATOR: Not excluding", obj]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)
        return False

    def _children(self, script: default.Script, obj: Atspi.Accessible) -> list:
        """Returns a list of children for obj, taking simple navigation into account."""

        if not AXObject.get_child_count(obj):
            return []

        children = list(AXObject.iter_children(obj))
        if not self._simplify:
            return children

        # Add the children of excluded objects to our list of children.
        functional_children = []
        for child in children:
            if self._exclude_from_simple_navigation(script, child):
                functional_children.extend(self._children(script, child))
            else:
                functional_children.append(child)

        return functional_children

    def _parent(self, script: default.Script, obj: Atspi.Accessible) -> Atspi.Accessible | None:
        """Returns the parent for obj, taking simple navigation into account."""

        parent = AXObject.get_parent(obj)
        if not self._simplify:
            return parent

        # The first non-excluded ancestor is the functional parent.
        while parent is not None and self._exclude_from_simple_navigation(script, parent):
            parent = AXObject.get_parent(parent)

        return parent

    def _set_navigator_focus(self, obj: Atspi.Accessible) -> None:
        """Changes the navigator focus, storing the previous focus."""

        self._last_navigator_focus = self._navigator_focus
        self._navigator_focus = obj

    def _update(self) -> None:
        """Updates the navigator focus to Orca's object of interest."""

        mode, region = focus_manager.get_manager().get_active_mode_and_object_of_interest()
        obj = region or focus_manager.get_manager().get_locus_of_focus()
        if self._last_locus_of_focus == obj \
           or (region is None and mode == focus_manager.FLAT_REVIEW):
            return

        self._navigator_focus = obj
        self._last_locus_of_focus = obj

    def _present(self, script: default.Script, notify_user: bool = True) -> None:
        """Presents the current navigator focus to the user."""

        tokens = ["OBJECT NAVIGATOR: Presenting", self._navigator_focus]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)
        focus_manager.get_manager().emit_region_changed(
            self._navigator_focus, mode=focus_manager.OBJECT_NAVIGATOR)
        if not notify_user:
            msg = "OBJECT NAVIGATOR: _present called with notify_user=False"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return

        script.present_object(self._navigator_focus, priorObj=self._last_navigator_focus)

    @dbus_service.command
    def move_to_parent(
        self,
        script: default.Script,
        event: input_event.InputEvent | None = None,
        notify_user: bool = True
    ) -> bool:
        """Moves the navigator focus to the parent of the current focus."""

        tokens = ["OBJECT NAVIGATOR: move_to_parent. Script:", script,
                  "Event:", event, "notify_user:", notify_user]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        self._update()
        parent = self._parent(script, self._navigator_focus)
        if parent is not None:
            self._set_navigator_focus(parent)
            self._present(script, notify_user)
        elif notify_user:
            script.present_message(messages.NAVIGATOR_NO_PARENT)
        return True

    @dbus_service.command
    def move_to_first_child(
        self,
        script: default.Script,
        event: input_event.InputEvent | None = None,
        notify_user: bool = True
    ) -> bool:
        """Moves the navigator focus to the first child of the current focus."""

        tokens = ["OBJECT NAVIGATOR: move_to_first_child. Script:", script,
                  "Event:", event, "notify_user:", notify_user]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        self._update()
        children = self._children(script, self._navigator_focus)
        if not children:
            if notify_user:
                script.present_message(messages.NAVIGATOR_NO_CHILDREN)
            return True

        self._set_navigator_focus(children[0])
        self._present(script, notify_user)
        return True

    @dbus_service.command
    def move_to_next_sibling(
        self,
        script: default.Script,
        event: input_event.InputEvent | None = None,
        notify_user: bool = True
    ) -> bool:
        """Moves the navigator focus to the next sibling of the current focus."""

        tokens = ["OBJECT NAVIGATOR: move_to_next_sibling. Script:", script,
                  "Event:", event, "notify_user:", notify_user]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        self._update()
        parent = self._parent(script, self._navigator_focus)
        if parent is None:
            if notify_user:
                script.present_message(messages.NAVIGATOR_NO_NEXT)
            return True

        siblings = self._children(script, parent)
        if self._navigator_focus in siblings:
            index = siblings.index(self._navigator_focus)
            if index < len(siblings) - 1:
                self._set_navigator_focus(siblings[index+1])
                self._present(script, notify_user)
            elif notify_user:
                script.present_message(messages.NAVIGATOR_NO_NEXT)
        else:
            self._set_navigator_focus(parent)
            self._present(script, notify_user)
        return True

    @dbus_service.command
    def move_to_previous_sibling(
        self,
        script: default.Script,
        event: input_event.InputEvent | None = None,
        notify_user: bool = True
    ) -> bool:
        """Moves the navigator focus to the previous sibling of the current focus."""

        tokens = ["OBJECT NAVIGATOR: move_to_previous_sibling. Script:", script,
                  "Event:", event, "notify_user:", notify_user]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        self._update()
        parent = self._parent(script, self._navigator_focus)
        if parent is None:
            if notify_user:
                script.present_message(messages.NAVIGATOR_NO_PREVIOUS)
            return True

        siblings = self._children(script, parent)
        if self._navigator_focus in siblings:
            index = siblings.index(self._navigator_focus)
            if index > 0:
                self._set_navigator_focus(siblings[index-1])
                self._present(script, notify_user)
            elif notify_user:
                script.present_message(messages.NAVIGATOR_NO_PREVIOUS)
        else:
            self._set_navigator_focus(parent)
            self._present(script, notify_user)
        return True

    @dbus_service.command
    def toggle_simplify(
        self,
        script: default.Script,
        event: input_event.InputEvent | None = None,
        notify_user: bool = True
    ) -> bool:
        """Toggles simplified navigation."""

        tokens = ["OBJECT NAVIGATOR: toggle_simplify. Script:", script,
                  "Event:", event, "notify_user:", notify_user]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        self._simplify = not self._simplify
        if notify_user:
            if self._simplify:
                script.present_message(messages.NAVIGATOR_SIMPLIFIED_ENABLED)
            else:
                script.present_message(messages.NAVIGATOR_SIMPLIFIED_DISABLED)
        return True

    @dbus_service.command
    def perform_action(
        self,
        script: default.Script,
        event: input_event.InputEvent | None = None,
        notify_user: bool = True
    ) -> bool:
        """Attempts to click on the current focus."""

        tokens = ["OBJECT NAVIGATOR: perform_action. Script:", script,
                  "Event:", event, "notify_user:", notify_user]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        if AXEventSynthesizer.try_all_clickable_actions(self._navigator_focus):
            return True

        AXEventSynthesizer.click_object(self._navigator_focus, 1)
        return True


_navigator: ObjectNavigator = ObjectNavigator()
def get_navigator() -> ObjectNavigator:
    """Returns the Object Navigator"""

    return _navigator



# Orca
#
# Copyright 2012 Igalia, S.L.
#
# Author: Joanmarie Diggs <jdiggs@igalia.com>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., Franklin Street, Fifth Floor,
# Boston MA  02110-1301 USA.

# pylint: disable=wrong-import-position
# pylint: disable=too-few-public-methods
# pylint: disable=too-many-locals

"""Displays a GUI for Orca navigation list dialogs"""

# This has to be the first non-docstring line in the module to make linters happy.
from __future__ import annotations

__id__        = "$Id:$"
__version__   = "$Revision:$"
__date__      = "$Date:$"
__copyright__ = "Copyright (c) 2012 Igalia, S.L."
__license__   = "LGPL"

import time
from typing import TYPE_CHECKING, Any

import gi
gi.require_version("Gdk", "3.0")
gi.require_version("Gtk", "3.0")
from gi.repository import Gdk, GObject, Gtk

from . import debug
from . import guilabels
from . import script_manager
from .ax_event_synthesizer import AXEventSynthesizer
from .ax_object import AXObject

if TYPE_CHECKING:
    gi.require_version("Atspi", "2.0")
    from gi.repository import Atspi

    from .scripts import default


class OrcaNavListGUI:
    """A GUI for displaying navigation lists in Orca."""

    def __init__(
        self,
        title: str,
        column_headers: list[str],
        rows: list[tuple[Any, ...]],
        selected_row: int
    ) -> None:
        """Initialize the navigation list GUI."""
        self._tree: Gtk.TreeView | None = None
        self._activate_button: Gtk.Button | None = None
        self._jump_to_button: Gtk.Button | None = None
        self._gui: Gtk.Dialog = self._create_nav_list_dialog(column_headers, rows, selected_row)
        self._gui.set_title(title)
        self._gui.set_modal(True)
        self._gui.set_keep_above(True) # pylint: disable=no-member
        self._gui.set_focus_on_map(True) # pylint: disable=no-member
        self._gui.set_accept_focus(True) # pylint: disable=no-member
        self._script: default.Script | None = script_manager.get_manager().get_active_script()
        self._document: Atspi.Accessible | None = None

    def _create_nav_list_dialog(
        self,
        column_headers: list[str],
        rows: list[tuple[Any, ...]],
        selected_row: int
    ) -> Gtk.Dialog:
        """Create the navigation list dialog."""
        dialog = Gtk.Dialog()
        dialog.set_default_size(500, 400)

        grid = Gtk.Grid()
        content_area = dialog.get_content_area()
        content_area.add(grid)

        scrolled_window = Gtk.ScrolledWindow()
        grid.add(scrolled_window) # pylint: disable=no-member

        self._tree = Gtk.TreeView()
        self._tree.set_hexpand(True)
        self._tree.set_vexpand(True)
        scrolled_window.add(self._tree) # pylint: disable=no-member

        cols: list[type] = [GObject.TYPE_OBJECT, GObject.TYPE_INT]
        cols.extend(len(column_headers) * [GObject.TYPE_STRING])
        model = Gtk.ListStore(*cols)

        cell = Gtk.CellRendererText()
        column = Gtk.TreeViewColumn("Accessible", cell, text=0)
        column.set_visible(False)
        self._tree.append_column(column)

        cell = Gtk.CellRendererText()
        column = Gtk.TreeViewColumn("offset", cell, text=1)
        column.set_visible(False)
        self._tree.append_column(column)

        for i, header in enumerate(column_headers):
            cell = Gtk.CellRendererText()
            column = Gtk.TreeViewColumn(header, cell, text=i + 2)
            column.set_sort_column_id(i + 2)
            self._tree.append_column(column)

        for row in rows:
            row_iter = model.append(None)
            for i, cell_value in enumerate(row):
                model.set_value(row_iter, i, cell_value)

        self._tree.set_model(model)
        selection = self._tree.get_selection()
        selection.select_path(selected_row)

        btn = dialog.add_button(guilabels.BTN_CANCEL, Gtk.ResponseType.CANCEL)
        btn.connect("clicked", self._on_cancel_clicked)

        self._jump_to_button = dialog.add_button(guilabels.BTN_JUMP_TO, Gtk.ResponseType.APPLY)
        self._jump_to_button.connect("clicked", self._on_jump_to_clicked)

        self._activate_button = dialog.add_button(guilabels.ACTIVATE, Gtk.ResponseType.OK)
        self._activate_button.connect("clicked", self._on_activate_clicked)
        self._activate_button.grab_default()

        self._tree.connect("key-release-event", self._on_key_release)
        self._tree.connect("cursor-changed", self._on_cursor_changed)
        self._tree.set_search_column(2)

        return dialog

    def show_gui(self) -> None:
        """Show the navigation list GUI."""
        self._gui.show_all() # pylint: disable=no-member
        self._gui.present_with_time(time.time())

    def _on_cursor_changed(self, _widget: Gtk.Widget) -> None:
        """Handle cursor change events in the tree view."""
        obj, _offset = self._get_selected_accessible_and_offset()
        n_actions = AXObject.get_n_actions(obj)
        if self._activate_button is not None:
            self._activate_button.set_sensitive(n_actions > 0)
            if n_actions > 0:
                self._activate_button.grab_default()
            elif self._jump_to_button is not None:
                self._jump_to_button.grab_default()

    def _on_key_release(self, _widget: Gtk.Widget, event: Gdk.EventKey) -> None:
        """Handle key release events in the tree view."""
        keycode = event.hardware_keycode
        keymap = Gdk.Keymap.get_default()
        entries_for_keycode = keymap.get_entries_for_keycode(keycode)
        entries = entries_for_keycode[-1]
        event_string = Gdk.keyval_name(entries[0])
        if event_string == "Return":
            self._gui.activate_default()

    def _on_cancel_clicked(self, _widget: Gtk.Widget) -> None:
        """Handle cancel button click."""
        self._gui.destroy()

    def _on_jump_to_clicked(self, _widget: Gtk.Widget) -> None:
        """Handle jump to button click."""
        obj, offset = self._get_selected_accessible_and_offset()
        self._gui.destroy()
        if self._script is not None:
            self._script.utilities.set_caret_position(obj, offset, self._document)

    def _on_activate_clicked(self, _widget: Gtk.Widget) -> None:
        """Handle activate button click."""
        obj, offset = self._get_selected_accessible_and_offset()
        self._gui.destroy()
        if not obj:
            return

        if self._script is not None:
            self._script.utilities.set_caret_position(obj, offset)
        if not AXEventSynthesizer.try_all_clickable_actions(obj):
            tokens = ["INFO: Attempting a synthesized click on", obj]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            AXEventSynthesizer.click_object(obj)

    def _get_selected_accessible_and_offset(self) -> tuple[Atspi.Accessible | None, int]:
        """Get the selected accessible object and offset."""
        if not self._tree:
            msg = "ERROR: Could not get navlist tree"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return None, -1

        selection = self._tree.get_selection()
        if not selection:
            msg = "ERROR: Could not get selection for navlist tree"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return None, -1

        model, paths = selection.get_selected_rows()
        if not paths:
            msg = "ERROR: Could not get paths for navlist tree"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return None, -1

        obj = model.get_value(model.get_iter(paths[0]), 0)
        offset = model.get_value(model.get_iter(paths[0]), 1)
        return obj, max(0, offset)


def show_ui(
    title: str = "",
    column_headers: list[str] | None = None,
    rows: list[tuple[Any, ...]] | None = None,
    selected_row: int = 0,
) -> None:
    """Show the navigation list UI."""
    if column_headers is None:
        column_headers = []
    if rows is None:
        rows = [()]

    gui = OrcaNavListGUI(title, column_headers, rows, selected_row)
    gui.show_gui()



# Orca
#
# Copyright 2014 Igalia, S.L.
#
# Author: Joanmarie Diggs <jdiggs@igalia.com>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., Franklin Street, Fifth Floor,
# Boston MA  02110-1301 USA.

"""Customized support for spellcheck in Thunderbird."""

# This has to be the first non-docstring line in the module to make linters happy.
from __future__ import annotations

__id__ = "$Id$"
__version__   = "$Revision$"
__date__      = "$Date$"
__copyright__ = "Copyright (c) 2014 Igalia, S.L."
__license__   = "LGPL"

from typing import TYPE_CHECKING

from orca import debug
from orca import spellcheck
from orca.ax_object import AXObject
from orca.ax_utilities import AXUtilities

if TYPE_CHECKING:
    import gi
    gi.require_version("Atspi", "2.0")
    from gi.repository import Atspi

class SpellCheck(spellcheck.SpellCheck):
    """Customized support for spellcheck in Thunderbird."""

    def _is_candidate_window(self, window: Atspi.Accessible) -> bool:
        """Returns True if window could be the spellcheck window pending other checks."""

        if not (AXUtilities.is_dialog(window) or AXUtilities.is_modal(window)):
            tokens = ["THUNDERBIRD SPELL CHECK:", window, "is not a dialog or modal window"]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return False

        def is_non_spell_check_child(x: Atspi.Accessible) -> bool:
            return AXUtilities.is_page_tab_list(x) or AXUtilities.is_split_pane(x)

        if AXObject.find_descendant(window, is_non_spell_check_child):
            return False

        return True
